
import java.util.*;
/**
 * Verwaltet die Fahrzeuge auf einer Fahrbahn.
 * 
 * @author Albert Wiedemann
 * @version 1.0
 */
class Fahrbahn extends Ereignisbehandlung
{
    /** x-Position */
    private int x;
    /** y-Position */
    private int y;
    /** start-Position */
    private int start;
    /** ende-Position */
    private int ende;
    /** rand-Position */
    private int rand;
    /** Wartezeit zur nächsten Fahrzeugerzeugung */
    private int wartezeit;
    
    /** die Fahrzeuge */
    private ArrayList<Fahrzeug> fahrzeuge;
    /** Darstellungssymbol */
    private FahrbahnSymbol darstellung;
    /** zuständige Ampel */
    private Ampel ampel;

    /**
     * Erzeugt eine leere Fahrbahn ab dem linken Rand unterhalb der Fenstermitte.
     * @param a für diese Richtung zuständige Ampel
     * @param x x-Koordinate der Fahrbahn
     * @param y y-Koordinate der Fahrbahn
     */
    Fahrbahn(Ampel a, int x, int y)
    {
        this.x = x;
        this.y = y;
        ampel = a;
        fahrzeuge = new ArrayList<Fahrzeug>();
        darstellung = new FahrbahnSymbol();
        darstellung.PositionSetzen(x, y);
        start = darstellung.EintrittsPositionGeben();
        ende = darstellung.AustrittsPositionGeben();
        rand = darstellung.RandAbstandGeben();
        wartezeit = 0;
    }

    /**
     * Reiht ein neues Fahrzeug an der Startposition in die Fahrbahn ein.
     * @param f das neue Fahrzeug
     */
    void FahrzeugEinreihen(Fahrzeug f)
    {
        fahrzeuge.add(f);
        f.PositionSetzen(x + rand, start);
    }

    /**
     * Ermittelt, ob ein Fahrzeug entfernt werden muss
     * @param f das in Frage kommende Fahrzeug
     * @return true, wenn das Fahrzeug die Fahrbahn verlassen hat
     */
    private boolean IstDraussen(Fahrzeug f)
    {
        return f.YKoordinateGeben() < ende;
    }

    /**
     * Ermittelt, ob das Fahrzeug sich nach dem Vorgänger richten oder
     * die Ampel beachten muss.
     * @param ampelabstand Abstand zur Ampel (Haltelinie)
     * @param vorgängerabstand Abstand zum Vorgängerfahrzeug
     * @param f das aktuelle Fahrzeug
     * @return true, wenn der Vorgänger wichtig ist, false, wenn die Ampel zählt
     */
    private boolean VorgängerZählt(int ampelabstand, int vorgängerabstand, Fahrzeug f)
    {
        return (ampelabstand < 0) || (vorgängerabstand < ampelabstand) || ampel.IstGrün() ||
               ampel.IstGelb() && (ampelabstand < f.GeschwindigkeitGeben() * 1.8 * 5) && (f.GeschwindigkeitGeben() > 3.0);
    }

    /**
     * Bewegt alle Fahrzeuge der Fahrbahn für eine Sekunde.
     */
    void FahrzeugeBewegen()
    {
        Fahrzeug vorgänger;
        double vvorgänger;
        int wegmax;
        int ampelAbstand;
        vorgänger = null;
        for (Fahrzeug f: fahrzeuge)
        {
            if (vorgänger == null)
            {
                wegmax = 100000;
                vvorgänger = 100;
            }
            else
            {
                vvorgänger = vorgänger.GeschwindigkeitGeben();
                wegmax = f.YKoordinateGeben() - vorgänger.YKoordinateGeben() - vorgänger.LängeGeben() - 2;
                if (vvorgänger > 0)
                {
                    wegmax = wegmax - (int) Math.round((f.GeschwindigkeitGeben() + vvorgänger) / 2.0 * 1.8 * 5);
                }
            }
            ampelAbstand = f.YKoordinateGeben() - darstellung.HaltelinienPositionGeben();
            if (VorgängerZählt (ampelAbstand, wegmax, f))
            {
                f.Fahren(wegmax, vvorgänger);
            }
            else
            {
                f.Fahren(ampelAbstand, 0.0);
            }
            vorgänger = f;
        }
        if ((fahrzeuge.size() > 0) && IstDraussen(fahrzeuge.get(0)))
        {
            fahrzeuge.get(0).Entfernen();
            fahrzeuge.remove(0);
        }
    }
    
    /**
     * Setzt die Position der Fahrbahn.
     * @param x x-Position
     * @param y y-Position
     */
    void PositionSetzen(int x, int y)
    {
        this.x = x;
        this.y = y;
        darstellung.PositionSetzen(x, y);
        rand = darstellung.RandAbstandGeben();
    }
    
    /**
     * Ausgabe der Fahrzeuganzahl
     * @return Anzahl der Fahrzeuge auf der Fahrbahn
     */
    int FahrzeuganzahlGeben()                  
    {
        return fahrzeuge.size();
    }

    /**
     * Erzeugt eine zufällige Farbe für das Fahrzeug
     * @return zufälliges Fahrzeug
     */
    private Fahrzeug FahrzeugErzeugen()
    {
        switch ((new java.util.Random()).nextInt(5))
        {
            case 0:
            case 1:
                return new PKW();
            case 2:
                return new Sportwagen();
            case 3:
                return new LKW();
            case 4:
                return new LKWMitAnhaenger();
            default:
                return null;
        }
    }

    /**
     * Fährt die Fahrzeuge und erzeugt gegebenenfalls ein neues Fahrzeug.
     */
    @Override void TaktImpulsAusführen()
    {
        FahrzeugeBewegen();
        if (wartezeit <= 0)
        {
            FahrzeugEinreihen(FahrzeugErzeugen());
            wartezeit = 4;
        }
        else
        {
            wartezeit = wartezeit - 1;
        }
    }
}
